# pylint: disable=E0611,C0103,R0914,W1514,R1735,R1734,C0206

import asyncio
from typing import Dict, List
from time import time
from zipfile import ZipFile
import tempfile
import json
import os
import shutil

import numpy as np
from ..pytpu import TPUDevice, TPUProgram, TPUInference, TPUProgramInfo, ProcessingMode  # type: ignore
from ..tools.helpers import to_raw
# from ..tools.helpers import STR_TO_BYTES
from ..tools.helpers import STR_TO_DTYPE
from ..tools.helpers import get_tpu_devices

__all__ = [
    'get_fps',
]


async def run_inference(name: str, device: TPUDevice, inference: TPUInference,
                        data_list: List[Dict[str, np.ndarray]]) -> None:

    mode = ProcessingMode.FULL

    for ii, data in enumerate(data_list):
        inference.load(data, mode=mode)
        status = await device.load_inference(inference)
        assert status.is_success
        inference.get(as_dict=True, mode=mode)

        if (ii + 1) % 10 == 0:
            print(f'{name}: finish {ii + 1} iteration')


def get_fps(program_path: str, raw: bool = False, n_queries: int = 100, n_proc: int = 4) -> float:

    print(f'Start measure performance for program: {program_path}')
    print(f'Configuration: RAW = {raw}; queries = {n_queries}; processes = {n_proc}')

    tpu_device_names = get_tpu_devices()

    with tempfile.TemporaryDirectory() as tempdir:
        with ZipFile(program_path, 'r') as zip_obj:
            zip_obj.extractall(tempdir)

        with open(os.path.join(tempdir, 'metadata.json'), 'r') as metadata_file:
            metadata = json.load(metadata_file)

        if raw is True:
            with open(os.path.join(tempdir, 'metadata.json'), 'w') as metadata_file:
                metadata = to_raw(metadata)
                json.dump(metadata, metadata_file)

            with tempfile.NamedTemporaryFile() as temp_file:
                program_path = os.path.join(tempdir, 'program_raw.tpu')
                shutil.make_archive(temp_file.name, 'zip', tempdir)
                os.rename(temp_file.name + '.zip', program_path)

            print(f'Raw program saved to {program_path}')

        layers_param = dict()
        for _, region in metadata['inputs'].items():
            for name, inp in region.items():
                layers_param[inp['anchor']] = inp

        data_list = list()
        for _ in range(n_queries):
            for name in layers_param:
                if raw:
                    data_shape = (1, layers_param[name]['size'], )
                    data_type = np.int8
                else:
                    data_shape = layers_param[name]['user_shape']
                    data_type = STR_TO_DTYPE[layers_param[name]['user_dtype']]

                generated_data = (np.random.rand(*data_shape) * 2 - 1) * 100
                generated_data = generated_data.astype(data_type)
                generated_data_dict = {name: generated_data}
                data_list.append(generated_data_dict)

        # input_name = list(layers_param.keys())[0]
        # program_name = os.path.basename(program_path)[:-4]
        # root_dir = os.path.dirname(program_path)
        # with open(os.path.join(root_dir, program_name + '_input.bin'), 'w') as f:
        #     data_list[0][input_name].tofile(f)

        batch = max([layers_param[name]['user_shape'][0] for name in layers_param])

        inference_processes = list()
        # tpu_device_names = ['/dev/tpu0', ]
        device_num = len(tpu_device_names)
        tpu_devices = list()
        for dev_ii, device_name in enumerate(tpu_device_names):
            tpu_devices.append(TPUDevice.open(device_name))
            tpu_program = TPUProgram(program_path, TPUProgramInfo(max_tasks_count=n_proc))
            tpu_devices[dev_ii].load_program(tpu_program)
            inferences = [TPUInference(tpu_program) for _ in range(n_proc)]
            inference_processes.extend([run_inference(
                name=f'Device {device_name} process {ii}',
                device=tpu_devices[dev_ii],
                inference=inferences[ii],
                data_list=data_list[ii:len(data_list):n_proc * device_num])
                                        for ii in range(n_proc)])

        start_time = time()
        asyncio.get_event_loop().run_until_complete(asyncio.gather(*inference_processes))
        total_inference_time = time() - start_time

        fps = n_queries * batch / total_inference_time

        print(f'Estimated FPS = {fps}')

    return fps



import importlib.util
import sys

spec = importlib.util.spec_from_file_location("libtpu", "/usr/lib/pytpu.so")
assert spec
libtpu = importlib.util.module_from_spec(spec)
sys.modules['pytpu.pytpu'] = libtpu


# pylint: disable=E0603
__all__ = [
    'ProcessingMode',
    'TPUDevice',
    'TPUInference',
    'TPUInferenceException',
    'TPUProgram',
    'TPUProgramException',
    'TPUProgramInfo',
]



import asyncio
import json
import logging
import os
import shutil
import tempfile

from math import ceil
from typing import Any
from typing import Dict
from typing import Tuple
from zipfile import ZipFile

import numpy as np

from tpu_tlm_is.base import TensorDescription
from tpu_tlm_is.base.number_types import UserNumberType
from tpu_tlm_is.base.number_types import TpuNumberType
from tpu_tlm_is.base.number_types import STR_TO_USER_NUMBER_TYPE
from tpu_tlm_is.base.number_types import STR_TO_TPU_NUMBER_TYPE
from tpu_tlm_is.models.iotools import post_process
from tpu_tlm_is.models.iotools import pre_process
from pytpu.pytpu import TPUDevice, TPUProgram, TPUInference  # type: ignore
# from pytpu.pytpu import ProcessingMode  # type: ignore
from ..tools.helpers import get_tpu_devices
from ..tools.helpers import get_tpu_parameters
from ..tools.helpers import to_raw

__all__ = [
    'TpuRunner',
]

LOGGER = logging.getLogger(__name__)
_SUFFIX = ':0'


def _get_tensor_description(io_: Dict[str, Any], cwl: int) -> TensorDescription:
    if 'user_shape' in io_.keys():
        LOGGER.debug('Generate NON-RAW descriptions')
        return TensorDescription(
            user_shape_mask=tuple(tuple([True, ] * abs(p[0]) + [False, ] * s + [True, ] * abs(p[1])
                                  for p, s in zip(io_['padding'], io_['user_shape']))),
            user_order=io_['user_order'],
            user_dtype=STR_TO_USER_NUMBER_TYPE[io_['user_dtype']],
            tpu_shape=io_['tpu_shape'],
            tpu_order=io_['tpu_order'],
            tpu_dtype=STR_TO_TPU_NUMBER_TYPE[io_['tpu_dtype']],
            scales=tuple([float(s) for s in io_['scales']]),
            anchor=io_['anchor'],
        )
    else:
        LOGGER.debug('Generate RAW descriptions')
        return TensorDescription(
            user_shape_mask=((False, ), tuple([False, ] * int(io_['size'])), ),
            user_order=('N', 'C', ),
            user_dtype=UserNumberType.INT8,
            tpu_shape=(1, ceil(int(io_['size']) / cwl), np.minimum(cwl, int(io_['size']))),
            tpu_order=('N', 'C', 'B'),
            tpu_dtype=TpuNumberType.INT8,
            scales=(1.0, ),
        )


def _tensor_as_node(node: Dict[str, np.ndarray]) -> Tuple[Dict[str, np.ndarray], bool]:
    node_tensor = {}
    _len_suffix = len(_SUFFIX)
    flag_suffix = False
    for key, data in node.items():
        if key[-_len_suffix:] == _SUFFIX:
            node_tensor[key[:-_len_suffix]] = data
            flag_suffix = True
        else:
            node_tensor[key] = data
    return node_tensor, flag_suffix


def _node_as_tensor(node: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
    node_tensor = {}
    _len_suffix = len(_SUFFIX)
    for key, data in node.items():
        if key[-_len_suffix:] == _SUFFIX:
            node_tensor[key] = data
        else:
            node_tensor[key + _SUFFIX] = data

    return node_tensor


class TpuRunner:
    """Runner for IVA TPU."""

    def __init__(self, program_path: str, loop: Any = None, sync: bool = False, pyprocessing: bool = True) -> None:

        LOGGER.debug(f'Init TpuRunner with program: {program_path}')

        self._pyprocessing = pyprocessing
        self._sync = sync
        self._loop = loop if loop else asyncio.get_event_loop()

        tpu_devices = get_tpu_devices()
        if len(tpu_devices) < 1:
            raise EnvironmentError('No TPU devices found')

        self._device = TPUDevice(tpu_devices[0])

        if pyprocessing:
            with tempfile.TemporaryDirectory() as tempdir:
                with ZipFile(program_path, 'r') as zip_obj:
                    zip_obj.extractall(tempdir)

                with open(os.path.join(tempdir, 'metadata.json'), 'r') as metadata_file:
                    metadata = json.load(metadata_file)

                tpu_par = get_tpu_parameters(metadata['hardware_parameters'])

                tensor_descriptions: Dict[str, TensorDescription] = dict()
                for _, region in metadata['inputs'].items():
                    for name, io_ in region.items():
                        tensor_descriptions[name] = _get_tensor_description(io_, tpu_par.cache_word_length)
                        LOGGER.debug(f'Input: {name}, {[len(s) for s in tensor_descriptions[name].user_shape_mask]}, '
                                     f'{tensor_descriptions[name].user_dtype}')

                for _, region in metadata['outputs'].items():
                    for name, io_ in region.items():
                        tensor_descriptions[name] = _get_tensor_description(io_, tpu_par.cache_word_length)

                with open(os.path.join(tempdir, 'metadata.json'), 'w') as metadata_file:
                    raw_metadata = to_raw(metadata)
                    json.dump(raw_metadata, metadata_file)

                with tempfile.NamedTemporaryFile() as temp_file:
                    program_path = os.path.join(tempdir, 'program_raw.tpu')
                    shutil.make_archive(temp_file.name, 'zip', tempdir)
                    os.rename(temp_file.name + '.zip', program_path)
                    LOGGER.debug(f'Raw program saved to {program_path}')

                self._tpu_par = tpu_par
                self._tensor_descriptions = tensor_descriptions
                self._program = TPUProgram(program_path)
        else:
            self._program = TPUProgram(program_path)

        self._device.load_program(self._program)

    def __call__(self, input_data: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
        """
        Run inference using IVA TPU.

        Parameters
        ----------
        input_data
            Dict with input tensors for inference

        """

        input_data, flag_suffix = _tensor_as_node(input_data)

        if self._pyprocessing:
            input_data = pre_process(self._tpu_par, input_data, self._tensor_descriptions)
            # Convert to raw program format
            input_data = {n: v.reshape((1, -1)).view(np.int8) for n, v in input_data.items()}

        inference = TPUInference(self._program)
        inference.load(input_data)

        if self._sync:
            self._device.load_inference_sync(inference)
        else:
            status_future = self._device.load_inference(inference)
            self._loop.run_until_complete(status_future)

        output_data = inference.get(as_dict=True)

        if self._pyprocessing:
            output_data = post_process(self._tpu_par, output_data, self._tensor_descriptions)

        if flag_suffix:
            output_data = _node_as_tensor(output_data)

        return output_data



import copy
import json
import logging
import os
import re
import shutil
import tempfile
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from zipfile import ZipFile

import numpy as np
from tpu_tlm_is.base import TpuParameters
from tpu_tlm_is.base import get_hw_params
from tpu_tlm_is.base.number_types import STR_TO_NUMPY


__all__ = [
    'get_tpu_devices',
    'get_tpu_parameters',
    'STR_TO_BYTES',
    'STR_TO_DTYPE',
    'to_raw',
    'to_nhwc_program',
]

LOGGER = logging.getLogger(__name__)


def get_tpu_devices() -> List[str]:
    return [os.path.join('/dev', f) for f in os.listdir('/dev') if re.match(r'tpu.', f) and len(f) == 4]


def get_tpu_parameters(hw_par: Dict[str, Any]) -> TpuParameters:
    if hw_par['cache_word_len'] == 128 and hw_par['ddr_word_len'] == 32:
        return get_hw_params('128x128').tpu_parameters
    if hw_par['cache_word_len'] == 64 and hw_par['ddr_word_len'] == 16:
        return get_hw_params('64x64').tpu_parameters
    if hw_par['cache_word_len'] == 64 and hw_par['ddr_word_len'] == 64:
        return get_hw_params('64x64_fpga').tpu_parameters

    raise NotImplementedError('Translation of such platform does not supported!')


def _to_raw_old(io_: Dict[str, Any], name: str) -> Dict[str, Any]:
    LOGGER.debug(f'IO description: {io_}')

    size = io_['size']
    io_out = {
        'address': io_['address'],
        'size': size,
        'user_shape': [1, size, ],
        'user_order': ['N', 'C', ],
        'user_dtype': 'int8',
        'tpu_shape': [1, size, ],
        'tpu_order': ['N', 'C', ],
        'tpu_dtype': 'int8',
        'padding': [[0, 0], [0, 0]],
        'scales': [1.0, ],
        'anchor': io_.get('anchor', name),
    }
    LOGGER.debug(f'RAW IO description: {io_out}')

    return io_out


def to_raw(metadata: Dict[str, Any]) -> Dict[str, Any]:
    raw_metadata = copy.copy(metadata)
    for idx, region in raw_metadata['inputs'].items():
        for name, _ in region.items():
            raw_metadata['inputs'][idx][name] = _to_raw_old(metadata['inputs'][idx][name], name)

    for idx, region in raw_metadata['outputs'].items():
        for name, _ in region.items():
            raw_metadata['outputs'][idx][name] = _to_raw_old(metadata['outputs'][idx][name], name)

    return raw_metadata


def _to_nhwc_descriptor(io_: Dict[str, Any], name: str) -> Dict[str, Any]:
    LOGGER.debug(f'IO description: {io_}')

    assert io_['user_order'][-1] == 'C'

    user_shape = io_['user_shape']
    user_order = io_['user_order']
    tpu_order = io_['tpu_order']
    tpu_order_replace = [user_order[-1], *user_order[1:-1], user_order[0]]
    tpu_shape = io_['tpu_shape']

    # print(tpu_shape)

    channels = tpu_shape[tpu_order.index('C')] * tpu_shape[tpu_order.index('B')]
    batch = tpu_shape[tpu_order.index('N')]
    tpu_shape_replace = [batch, *tpu_shape[1:-2], channels]

    # print(tpu_shape_replace)

    size = io_['size']
    io_out = {
        'address': io_['address'],
        'size': size,
        'user_shape': user_shape,
        'user_order': user_order,
        'user_dtype': io_['user_dtype'],
        'tpu_shape': tpu_shape_replace,
        'tpu_order': tpu_order_replace,
        'tpu_dtype': io_['tpu_dtype'],
        'padding': io_['padding'],
        'scales': io_['scales'],
        'anchor': io_.get('anchor', name),
    }
    LOGGER.debug(f'RAW IO description: {io_out}')

    return io_out


def to_nhwc_metadata(metadata: Dict[str, Any]) -> Dict[str, Any]:
    nhwc_metadata = copy.copy(metadata)
    for idx, region in nhwc_metadata['inputs'].items():
        for name, _ in region.items():
            nhwc_metadata['inputs'][idx][name] = _to_nhwc_descriptor(metadata['inputs'][idx][name], name)

    for idx, region in nhwc_metadata['outputs'].items():
        for name, _ in region.items():
            nhwc_metadata['outputs'][idx][name] = _to_nhwc_descriptor(metadata['outputs'][idx][name], name)

    return nhwc_metadata


def to_nhwc_program(source_path: str, store_path: str) -> None:
    with tempfile.TemporaryDirectory() as tempdir:
        with ZipFile(source_path, 'r') as zip_obj:
            zip_obj.extractall(tempdir)

        with open(os.path.join(tempdir, 'metadata.json'), 'r') as metadata_file:
            metadata = json.load(metadata_file)

        with open(os.path.join(tempdir, 'metadata.json'), 'w') as metadata_file:
            nhwc_metadata = to_nhwc_metadata(metadata)
            json.dump(nhwc_metadata, metadata_file)

        with tempfile.NamedTemporaryFile() as temp_file:
            shutil.make_archive(temp_file.name, 'zip', tempdir)
            os.rename(temp_file.name + '.zip', store_path)
            LOGGER.debug(f'NHWC program saved to {store_path}')


def get_io_from_bin(program_path: str, io_paths: Optional[List[str]], io_type: str
                    ) -> Tuple[Dict[str, np.ndarray], List[str]]:
    with tempfile.TemporaryDirectory() as tempdir:
        with ZipFile(program_path, 'r') as zip_obj:
            zip_obj.extractall(tempdir)

        with open(os.path.join(tempdir, 'metadata.json'), 'r', encoding='utf8') as metadata_file:
            metadata = json.load(metadata_file)

    io_data = {}
    io_names = []
    idx = 0
    for _, region in metadata[io_type].items():
        for name, io_ in region.items():
            io_names.append(name)
            if io_paths is not None:
                with open(io_paths[idx], 'r') as io_file:
                    tmp = np.fromfile(io_file, STR_TO_NUMPY[io_.get('user_dtype', 'int8')])
                    io_data[name] = tmp.reshape(io_.get('user_shape', (1, -1)))
                    LOGGER.debug(f'Read {io_type} from files: {name}, {io_data[name].shape}, {io_data[name].dtype}')
            idx += 1

    return io_data, io_names


STR_TO_DTYPE = {
    'int8': np.int8,
    'float16': np.float16,
    'float32': np.float32,
}

STR_TO_BYTES = {
    'int8': 1,
    'float16': 2,
    'float32': 4,
}
